# -*- coding: utf-8 -*-

from __future__ import print_function

import functools
import logging

from django.conf import settings
from django.core import mail
from django.utils import translation
try:
    # Django >= 1.4.5
    from django.utils.six import string_types
except ImportError:
    # Django < 1.4.5
    string_types = basestring
from django.template import loader, TemplateDoesNotExist

from . import models
from . import exceptions as exc


log = logging.getLogger("djmail")


def _get_body_template_prototype():
    return getattr(settings, "DJMAIL_BODY_TEMPLATE_PROTOTYPE",
                   "emails/{name}-body-{type}.{ext}")


def _get_subject_template_prototype():
    return getattr(settings, "DJMAIL_SUBJECT_TEMPLATE_PROTOTYPE",
                   "emails/{name}-subject.{ext}")


def _get_template_extension():
    return getattr(settings, "DJMAIL_TEMPLATE_EXTENSION", "html")


def _trap_exception(function):
    """
    Simple decorator for catch template exceptions. If exception is throwed,
    this decorator by default returns an empty string.
    """
    @functools.wraps(function)
    def _decorator(*args, **kwargs):
        try:
            return function(*args, **kwargs)
        except TemplateDoesNotExist as e:
            log.warning("Template '{0}' does not exists.".format(e))
            return u""

    return _decorator


def _trap_language(function):
    """
    Decorator that intercept a language attribute and set it on context of
    the execution.
    """
    @functools.wraps(function)
    def _decorator(self, ctx):
        language_new = None
        language_old = translation.get_language()

        # If attr found on context, set a new language
        if "lang" in ctx:
            language_new = ctx["lang"]
            translation.activate(language_new)

        try:
            return function(self, ctx)
        finally:
            if language_new is not None:
                translation.activate(language_old)

    return _decorator


class TemplateMail(object):
    name = None

    def __init__(self, name=None):
        self._email = None
        if name is not None:
            self.name = name

        self._initialize_settings()

    def _initialize_settings(self):
        self._body_template_name = _get_body_template_prototype()
        self._subject_template_name = _get_subject_template_prototype()

    @_trap_exception
    @_trap_language
    def _render_message_body_as_html(self, ctx):
        template_ext = _get_template_extension()
        template_name = self._body_template_name.format(**{
            "ext": template_ext, "name": self.name, "type":"html"})

        return loader.render_to_string(template_name, ctx)

    @_trap_exception
    @_trap_language
    def _render_message_body_as_txt(self, ctx):
        template_ext = _get_template_extension()
        template_name = self._body_template_name.format(**{
            "ext": template_ext, "name": self.name, "type": "text"})

        return loader.render_to_string(template_name, ctx)

    def _render_message_subject(self, ctx):
        template_ext = _get_template_extension()
        template_name = self._subject_template_name.format(**{
            "ext": template_ext, "name": self.name})

        try:
            subject = loader.render_to_string(template_name, ctx)
        except TemplateDoesNotExist as e:
            raise exc.TemplateNotFound("Template '{0}' does not exists.".format(e))
        return u" ".join(subject.strip().split())

    def _attach_body_to_email_instance(self, email, ctx):
        body_html = self._render_message_body_as_html(ctx)
        body_txt = self._render_message_body_as_txt(ctx)

        if not body_txt and not body_html:
            raise exc.TemplateNotFound("Body of email message shouldn't be empty")

        if isinstance(email, mail.EmailMultiAlternatives):
            email.body = body_txt
            email.attach_alternative(body_html, "text/html")
        else:
            email.body = body_txt

    def _attach_subject_to_email_instance(self, email, ctx):
        email.subject = self._render_message_subject(ctx)

    def make_email_object(self, to, ctx, **kwargs):
        if not isinstance(to, (list, tuple)):
            to = [to]

        email = mail.EmailMultiAlternatives(**kwargs)
        email.to = to

        self._attach_body_to_email_instance(email, ctx)
        self._attach_subject_to_email_instance(email, ctx)
        return email

    def send(self, to, context, **kwargs):
        email = self.make_email_object(to, context, **kwargs)
        return email.send()


class MagicMailBuilder(object):
    def __init__(self, email_attr="email", lang_attr="lang",
                 template_mail_cls=TemplateMail):
        self._email_attr = email_attr
        self._lang_attr = lang_attr
        self._template_mail_cls = template_mail_cls

    def __getattr__(self, name):
        def _dynamic_email_generator(to, ctx, priority=models.PRIORITY_STANDARD):
            lang = None

            if not isinstance(to, string_types):
                if not hasattr(to, self._email_attr):
                    raise AttributeError(
                        "'to' parameter does not "
                        "have '{0}' attribute".format(self._email_attr))

                lang = getattr(to, self._lang_attr, None)
                to = getattr(to, self._email_attr)

            if lang is not None:
                ctx["lang"] = lang

            template_email = self._template_mail_cls(name=name)
            email_instance = template_email.make_email_object(to, ctx)
            email_instance.priority = priority

            return email_instance

        return _dynamic_email_generator



# -*- coding: utf-8 -*-

import io

from django.conf import settings
from django.core.mail import get_connection
from django.core.paginator import Paginator
from django.utils import timezone
import traceback

from . import models


def _chunked_iterate_queryset(queryset, chunk_size=10):
    """
    Given a queryset, use paginator for iterate over all queryset
    but obtaining from database delimeted set of result parametrized
    with `chunk_size` parameter.
    """
    paginator = Paginator(queryset, chunk_size)
    for page_index in paginator.page_range:
        page = paginator.page(page_index)
        for item in page.object_list:
            yield item

def _safe_send_message(message_model, connection):
    """
    Given a message model, try send it, if send process
    is fail, increment retry count and save stack trace
    in message model.
    """
    email = message_model.get_email_message()
    sended = 0

    with io.StringIO() as file:
        try:
            sended = connection.send_messages([email])
        except Exception as e:
            traceback.print_exc(file=file)
            file.seek(0)
            message_model.exception = file.read()
        else:
            if sended == 1:
                message_model.status = models.STATUS_SENT
                message_model.sent_at = timezone.now()
            else:
                message_model.status = models.STATUS_FAILED
                message_model.retry_count += 1

        message_model.save()
        return sended

def _get_real_backend():
    real_backend_path = getattr(settings, "DJMAIL_REAL_BACKEND",
                           'django.core.mail.backends.console.EmailBackend')
    return get_connection(backend=real_backend_path, fail_silently=False)

def _send_messages(email_messages):
    connection = _get_real_backend()

    # Create a messages on a database for correct
    # tracking of their status.
    email_models = [models.Message.from_email_message(email, save=True)
                    for email in email_messages]

    # Open connection for send all messages
    connection.open()

    sended_counter = 0
    for email, model_instance in zip(email_messages, email_models):
        if hasattr(email, "priority"):
            if email.priority <= models.PRIORITY_LOW:
                model_instance.priority = email.priority
                model_instance.status = models.STATUS_PENDING
                model_instance.save()
                sended_counter += 1
                continue

        sended_counter += _safe_send_message(model_instance, connection)

    connection.close()
    return sended_counter

def _send_pending_messages():
    """
    Function that sends pending, low priority messages.
    """
    max_retry_value = getattr(settings, "DJMAIL_MAX_RETRY_NUMBER", 3)
    queryset = models.Message.objects.filter(status=models.STATUS_PENDING)\
                                     .order_by("-priority", "created_at")
    connection = _get_real_backend()
    connection.open()

    sended_counter = 0
    for message_model in _chunked_iterate_queryset(queryset, 100):
        # Use one unique connection for send all messages
        sended_counter += _safe_send_message(message_model, connection)

    connection.close()
    return sended_counter

def _retry_send_messages():
    """
    Function that retry send failed messages.
    """
    max_retry_value = getattr(settings, "DJMAIL_MAX_RETRY_NUMBER", 3)
    queryset = models.Message.objects.filter(status=models.STATUS_FAILED)\
                                     .filter(retry_count__lte=max_retry_value)\
                                     .order_by("-priority", "created_at")

    connection = _get_real_backend()
    connection.open()

    sended_counter = 0
    for message_model in _chunked_iterate_queryset(queryset, 100):
        sended_counter += _safe_send_message(message_model, connection)

    connection.close()
    return sended_counter

def _mark_discarded_messages():
    """
    Function that search messaged that exceds the global retry
    number and marks its as discarded messages.
    """

    max_retry_value = getattr(settings, "DJMAIL_MAX_RETRY_NUMBER", 3)
    queryset = models.Message.objects.filter(status=models.STATUS_FAILED,
                                             retry_count__gt=max_retry_value)
    return queryset.update(status=models.STATUS_DISCARDED)



# -*- coding: utf-8 -*-

from celery.task import task
from . import core


@task(name="tasks.send_messages")
def send_messages(messages):
    """
    Celery standard task for send async messages.
    """
    return core._send_messages(messages)


@task(name="tasks.retry_send_messages")
def retry_send_messages():
    """
    Celery periodic task for retry send failed messages.
    """
    core._send_pending_messages()
    core._mark_discarded_messages()
    core._retry_send_messages()



import threading
import functools
from concurrent.futures import Future, ThreadPoolExecutor

from django.db import connection
from django.conf import settings
from djmail import core

from . import base

# TODO: parametrize this
executor = ThreadPoolExecutor(max_workers=1)


def _close_connection_on_finish(function):
    """
    Decorator for future task, that closes
    django database connection when it ends.
    """
    @functools.wraps(function)
    def _decorator(*args, **kwargs):
        try:
            return function(*args, **kwargs)
        finally:
            connection.close()

    return _decorator


class EmailBackend(base.BaseEmailBackend):
    """
    djmail async backend that uses threadpool
    for send emails instead of other async task
    libraries like celery.
    """
    def _send_messages(self, email_messages):
        if len(email_messages) == 0:
            future = Future()
            future.set_result(0)
            return future

        @_close_connection_on_finish
        def _send(messages):
            return core._send_messages(email_messages)

        return executor.submit(_send, email_messages)



# -*- coding: utf-8 -*-


class BaseEmailBackend(object):
    """
    Base class that implements a django
    mail backend interface.
    """

    def __init__(self, *args, **kwargs):
        self.args = args
        self.kwargs = kwargs

    def open(self):
        pass

    def close(self):
        pass

    def _send_messages(self, email_messages):
        raise NotImplementedError

    def send_messages(self, email_messages):
        return self._send_messages(email_messages)




# -*- coding: utf-8 -*-

from django.core.mail import EmailMessage
from django.core import mail
from django.db import connection
from django.test import TestCase
from django.test.utils import override_settings

from . import models
from . import core
from .template_mail import TemplateMail, MagicMailBuilder


class TestEmailSending(TestCase):
    def setUp(self):
        models.Message.objects.all().delete()

    @override_settings(
        EMAIL_BACKEND="djmail.backends.default.EmailBackend",
        DJMAIL_REAL_BACKEND="django.core.mail.backends.locmem.EmailBackend")
    def test_simple_send_email(self):
        email = EmailMessage('Hello', 'Body goes here', 'from@example.com',
                             ['to1@example.com', 'to2@example.com'])

        email.send()
        self.assertEqual(len(mail.outbox), 1)
        self.assertEqual(models.Message.objects.count(), 1)

    @override_settings(
        EMAIL_BACKEND="djmail.backends.async.EmailBackend",
        DJMAIL_REAL_BACKEND="django.core.mail.backends.locmem.EmailBackend",
        DJMAIL_SEND_ASYNC=True)
    def test_async_send_email(self):
        email = EmailMessage('Hello', 'Body goes here', 'from@example.com',
                             ['to1@example.com', 'to2@example.com'])
        # If async is activated, send method
        # returns a thread instead of a number of
        # sent messages.
        future = email.send()
        result = future.result()

        self.assertEqual(len(mail.outbox), 1)
        self.assertEqual(models.Message.objects.count(), 1)

    @override_settings(
        EMAIL_BACKEND="djmail.backends.default.EmailBackend",
        DJMAIL_REAL_BACKEND="testing.mocks.BrokenEmailBackend")
    def test_failing_simple_send_email(self):
        email = EmailMessage('Hello', 'Body goes here', 'from@example.com',
                             ['to1@example.com', 'to2@example.com'])

        number_sent_emails = email.send()

        self.assertEqual(number_sent_emails, 0)
        self.assertEqual(len(mail.outbox), 0)
        self.assertEqual(models.Message.objects.count(), 1)

        mailmodel = models.Message.objects.get()
        self.assertEqual(mailmodel.status, models.STATUS_FAILED)

    @override_settings(
        EMAIL_BACKEND="djmail.backends.async.EmailBackend",
        DJMAIL_REAL_BACKEND="testing.mocks.BrokenEmailBackend")
    def test_failing_async_send_email(self):
        email = EmailMessage('Hello', 'Body goes here', 'from@example.com',
                             ['to1@example.com', 'to2@example.com'])

        f = email.send()
        rs = f.result()

        self.assertEqual(len(mail.outbox), 0)
        self.assertEqual(models.Message.objects.count(), 1)

        mailmodel = models.Message.objects.get()
        self.assertEqual(mailmodel.status, models.STATUS_FAILED)

    @override_settings(
        EMAIL_BACKEND="djmail.backends.celery.EmailBackend",
        DJMAIL_REAL_BACKEND="django.core.mail.backends.locmem.EmailBackend")
    def test_async_send_email_with_celery(self):
        email = EmailMessage('Hello', 'Body goes here', 'from@example.com',
                             ['to1@example.com', 'to2@example.com'])
        result = email.send()
        result.wait()

        self.assertEqual(len(mail.outbox), 1)
        self.assertEqual(models.Message.objects.count(), 1)

    @override_settings(
        EMAIL_BACKEND="djmail.backends.celery.EmailBackend",
        DJMAIL_REAL_BACKEND="testing.mocks.BrokenEmailBackend")
    def test_failing_async_send_email_with_celery(self):
        email = EmailMessage('Hello', 'Body goes here', 'from@example.com',
                             ['to1@example.com', 'to2@example.com'])
        result = email.send()
        result.wait()

        self.assertEqual(len(mail.outbox), 0)
        self.assertEqual(models.Message.objects.count(), 1)

        mailmodel = models.Message.objects.get()
        self.assertEqual(mailmodel.status, models.STATUS_FAILED)

    @override_settings(
        EMAIL_BACKEND="djmail.backends.celery.EmailBackend",
        DJMAIL_REAL_BACKEND="testing.mocks.BrokenEmailBackend")
    def test_failing_retry_send_01(self):
        email = EmailMessage('Hello', 'Body goes here', 'from@example.com',
                             ['to1@example.com', 'to2@example.com'])
        message_model = models.Message.from_email_message(email)
        message_model.status = models.STATUS_FAILED
        message_model.retry_count = 1
        message_model.save()

        core._retry_send_messages()

        message_model_2 = models.Message.objects.get(pk=message_model.pk)
        self.assertEqual(message_model_2.retry_count, 2)

    @override_settings(
        EMAIL_BACKEND="djmail.backends.celery.EmailBackend",
        DJMAIL_REAL_BACKEND="testing.mocks.BrokenEmailBackend",
        DJMAIL_MAX_RETRY_NUMBER=2)
    def test_failing_retry_send_01(self):
        email = EmailMessage('Hello', 'Body goes here', 'from@example.com',
                             ['to1@example.com', 'to2@example.com'])
        message_model = models.Message.from_email_message(email)
        message_model.status = models.STATUS_FAILED
        message_model.retry_count = 3
        message_model.save()

        core._mark_discarded_messages()

        message_model_2 = models.Message.objects.get(pk=message_model.pk)
        self.assertEqual(message_model_2.retry_count, 3)
        self.assertEqual(message_model_2.status, models.STATUS_DISCARDED)


class TestTemplateEmailSending(TestCase):
    def setUp(self):
        models.Message.objects.all().delete()

    @override_settings(
        EMAIL_BACKEND="djmail.backends.default.EmailBackend",
        DJMAIL_REAL_BACKEND="django.core.mail.backends.locmem.EmailBackend")
    def test_simple_send_email_1(self):
        class SimpleTemplateMail(TemplateMail):
            name = "test_email1"

        email = SimpleTemplateMail()
        email.send("to@example.com", {"name": "foo"})

        self.assertEqual(len(mail.outbox), 1)
        self.assertEqual(models.Message.objects.count(), 1)

        m = mail.outbox[0]
        self.assertEqual(m.subject, u'Subject1: foo')
        self.assertEqual(m.alternatives[0][0], u"<b>Mail1: foo</b>\n")

    @override_settings(
        EMAIL_BACKEND="djmail.backends.default.EmailBackend",
        DJMAIL_REAL_BACKEND="django.core.mail.backends.locmem.EmailBackend")
    def test_simple_send_email_2(self):
        class SimpleTemplateMail(TemplateMail):
            name = "test_email2"

        email = SimpleTemplateMail()
        email.send("to@example.com", {"name": "foo"})

        self.assertEqual(len(mail.outbox), 1)
        self.assertEqual(models.Message.objects.count(), 1)

        m = mail.outbox[0]
        self.assertEqual(m.subject, u'Subject2: foo')
        self.assertEqual(m.body, u"body\n")
        self.assertEqual(m.alternatives, [(u'<b>Body</b>\n', 'text/html')])

    @override_settings(
        EMAIL_BACKEND="djmail.backends.default.EmailBackend",
        DJMAIL_REAL_BACKEND="django.core.mail.backends.locmem.EmailBackend")
    def test_simple_send_email_with_magic_builder_1(self):
        mails = MagicMailBuilder()

        email = mails.test_email2("to@example.com", {"name": "foo"});
        email.send()

        self.assertEqual(len(mail.outbox), 1)
        self.assertEqual(models.Message.objects.count(), 1)

        self.assertEqual(email.subject, u'Subject2: foo')
        self.assertEqual(email.body, u"body\n")
        self.assertEqual(email.alternatives, [(u'<b>Body</b>\n', 'text/html')])


    @override_settings(
        EMAIL_BACKEND="djmail.backends.default.EmailBackend",
        DJMAIL_REAL_BACKEND="django.core.mail.backends.locmem.EmailBackend")
    def test_simple_send_email_with_magic_builder_1_with_low_priority(self):
        mails = MagicMailBuilder()

        email = mails.test_email2("to@example.com", {"name": "foo"}, priority=10);
        email.send()

        self.assertEqual(len(mail.outbox), 0)
        self.assertEqual(models.Message.objects.count(), 1)


        m1 = models.Message.objects.get()
        self.assertEqual(m1.status, models.STATUS_PENDING)
        self.assertEqual(m1.priority, 10)

        core._send_pending_messages()

        self.assertEqual(len(mail.outbox), 1)
        self.assertEqual(models.Message.objects.count(), 1)

        m2 = models.Message.objects.get()
        self.assertEqual(m2.status, models.STATUS_SENT)
        self.assertEqual(m2.priority, 10)


class SerializationEmailTests(TestCase):
    def setUp(self):
        models.Message.objects.all().delete()

    @override_settings(
        EMAIL_BACKEND="djmail.backends.default.EmailBackend",
        DJMAIL_REAL_BACKEND="django.core.mail.backends.locmem.EmailBackend")
    def test_simple_send_email_with_magic_builder_1(self):
        mails = MagicMailBuilder()

        email = mails.test_email2("to@example.com", {"name": "foo"});
        email.send()

        model = models.Message.objects.get()
        self.assertEqual(email.from_email, model.from_email)
        self.assertEqual(email.to, model.to_email.split(","))
        self.assertEqual(email.subject, model.subject)
        self.assertEqual(email.body, model.body_text)

        # self.assertEqual(len(mail.outbox), 1)
        # self.assertEqual(models.Message.objects.count(), 1)

        # self.assertEqual(email.subject, u'Subject2: foo')
        # self.assertEqual(email.body, u"body\n")
        # self.assertEqual(email.alternatives, [(u'<b>Body</b>\n', 'text/html')])




# -*- coding: utf-8 -*-

from __future__ import unicode_literals

import base64
import pickle
import uuid

try:
    # Django >= 1.4.5
    from django.utils.encoding import force_bytes, force_text
except ImportError:
    # Django < 1.4.5
    from django.utils.encoding import (
        smart_unicode as force_text, smart_str as force_bytes)
from django.db import models


STATUS_DRAFT = 10
STATUS_PENDING = 20
STATUS_SENT = 30
STATUS_FAILED = 40
STATUS_DISCARDED = 50

PRIORITY_LOW = 20
PRIORITY_STANDARD = 50


class Message(models.Model):
    STATUS_CHOICES = (
        (STATUS_DRAFT, "Draft"),
        (STATUS_SENT, "Sent"),
        (STATUS_FAILED, "Failed"),
        (STATUS_DISCARDED, "Discarded"),
    )

    uuid = models.CharField(max_length=40, primary_key=True,
                            default=lambda: str(uuid.uuid1()))

    from_email = models.CharField(max_length=1024, blank=True)
    to_email = models.TextField(blank=True)
    body_text = models.TextField(blank=True)
    body_html = models.TextField(blank=True)
    subject = models.CharField(max_length=1024, blank=True)

    data = models.TextField(blank=True, editable=False)

    retry_count = models.SmallIntegerField(default=-1)
    status = models.SmallIntegerField(choices=STATUS_CHOICES, default=STATUS_DRAFT)
    priority = models.SmallIntegerField(default=PRIORITY_STANDARD)

    created_at = models.DateTimeField(auto_now_add=True)
    sent_at = models.DateTimeField(null=True, default=None)
    exception = models.TextField(editable=True, blank=True)

    def get_email_message(self):
        raw_pickle_data = base64.b64decode(force_bytes(self.data))
        return pickle.loads(raw_pickle_data)

    @classmethod
    def from_email_message(cls, email_message, save=False):
        kwargs = {
            "from_email": force_text(email_message.from_email),
            "to_email": ",".join(force_text(x) for x in email_message.to),
            "body_text": force_text(email_message.body),
            "subject": force_text(email_message.subject),
            "data": base64.b64encode(pickle.dumps(email_message)),
        }

        if hasattr(email_message, "alternatives") and len(email_message.alternatives) > 0:
            html_alts = [x[0] for x in email_message.alternatives if x[1] == "text/html"]
            kwargs["body_html"] = html_alts[0] if len(html_alts) > 0 else ""

        instance = cls(**kwargs)
        if save:
            instance.save()

        return instance

    class Meta:
        ordering = ["created_at"]
        verbose_name = "Message"
        verbose_name_plural = "Messages"



# -*- coding: utf-8 -*-

from django.contrib import admin
from . import models


class MessageAdmin(admin.ModelAdmin):
    list_display = ('uuid', 'status', 'priority', 'created_at', 'sent_at', 'retry_count')
    list_display_links = list_display

    list_filter = ('status', 'priority', 'created_at', 'sent_at', 'retry_count')

admin.site.register(models.Message, MessageAdmin)



# -*- coding: utf-8 -*-

from django.conf import settings
from djmail import core

from . import base


class EmailBackend(base.BaseEmailBackend):
    def _send_messages(self, email_messages):
        if len(email_messages) == 0:
            return 0

        return core._send_messages(email_messages)



# -*- coding: utf-8 -*-

from __future__ import absolute_import

from django.conf import settings
from djmail.backends import base
from djmail import core
from djmail import tasks


class EmailBackend(base.BaseEmailBackend):
    """
    djmail backend that uses celery task for
    send emails.
    """
    def _send_messages(self, email_messages):
        if len(email_messages) == 0:
            return 0

        return tasks.send_messages.delay(email_messages)



from setuptools import setup

description = """
Simple, powerfull and nonobstructive django email middleware.
"""

setup(
    name = "djmail",
    url = "https://github.com/niwibe/djmail",
    author = "Andrey Antukh",
    author_email = "niwi@niwi.be",
    version="0.7",
    packages = [
        "djmail",
        "djmail.backends",
        "djmail.management",
        "djmail.management.commands",
    ],
    description = description.strip(),
    zip_safe=False,
    include_package_data = True,
    package_data = {
        "": ["*.html"],
    },
    classifiers = [
        # "Development Status :: 5 - Production/Stable",
        "Development Status :: 4 - Beta",
        "Operating System :: OS Independent",
        "Environment :: Web Environment",
        "Framework :: Django",
        "License :: OSI Approved :: BSD License",
        "Intended Audience :: Developers",
        "Programming Language :: Python",
        "Programming Language :: Python :: 2",
        "Programming Language :: Python :: 2.7",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.2",
        "Programming Language :: Python :: 3.3",
        "Topic :: Software Development :: Libraries",
        "Topic :: Utilities",
    ],
)



# -*- coding: utf-8 -*-

from django.core.management.base import NoArgsCommand, CommandError
from djmail import core

class Command(NoArgsCommand):
    def handle_noargs(**options):
        core._send_pending_messages()
        core._mark_discarded_messages()
        core._retry_send_messages()

        return 0
